#include <iostream>
#include <cmath>              // mathematical C library
#include <algorithm>             // defining the max() operator
#include <vector>           // STL vector templates
using namespace std;
#include "finrecipes.h"


#ifndef PI 
const double PI=3.141592653589793238462643;
#endif 

using namespace std;

double option_price_call_european_finite_diff_explicit(const double& S, 
						       const double& K, 
						       const double& r,
						       const double& sigma,
						       const double& time,
						       const int& no_S_steps,
						       const int& no_t_steps) {
    double sigma_sqr = pow(sigma,2);
    int M=no_S_steps; if ((no_S_steps%2)==1) { ++M; } // need no_S_steps to be even:
    double delta_S = 2.0*S/M;
    vector<double> S_values(M+1);
    for (int m=0;m<=M;m++) { S_values[m] = m*delta_S; };
    int N=no_t_steps;
    double delta_t = time/N;
    
    vector<double> a(M);
    vector<double> b(M);
    vector<double> c(M);
    double r1=1.0/(1.0+r*delta_t);
    double r2=delta_t/(1.0+r*delta_t);
    for (unsigned int j=1;j<M;j++){
	a[j] = r2*0.5*j*(-r+sigma_sqr*j);
	b[j] = r1*(1.0-sigma_sqr*j*j*delta_t);
	c[j] = r2*0.5*j*(r+sigma_sqr*j);
    };
    vector<double> f_next(M+1);
    for (int m=0;m<=M;++m) { f_next[m]=max(0.0,S_values[m]-K); };
    vector<double> f(M+1);
    for (int t=N-1;t>=0;--t) {
	f[0]=0;
	for (unsigned m=1;m<M;++m) {
	    f[m]=a[m]*f_next[m-1]+b[m]*f_next[m]+c[m]*f_next[m+1];
	};
	f[M] = 0;
	for (unsigned m=0;m<=M;++m) { f_next[m] = f[m]; };
    };
    return f[M/2];
};

double option_price_put_american_finite_diff_explicit( const double&  S, 
						       const double&  K, 
						       const double&  r,
						       const double&  sigma,
						       const double&  time,
						       const int& no_S_steps,
						       const int& no_t_steps) {
    double sigma_sqr = sigma*sigma;
    int M=no_S_steps+(no_S_steps%2); // need no_S_steps to be even:
    double delta_S = 2.0*S/M;
    vector<double> S_values(M+1);
    for (int m=0;m<=M;m++) { S_values[m] = m*delta_S; };
    int N=no_t_steps;
    double delta_t = time/N;
    
    vector<double> a(M);
    vector<double> b(M);
    vector<double> c(M);
    double r1=1.0/(1.0+r*delta_t);
    double r2=delta_t/(1.0+r*delta_t);
    for (int j=1;j<M;j++){
	a[j] = r2*0.5*j*(-r+sigma_sqr*j);
	b[j] = r1*(1.0-sigma_sqr*j*j*delta_t);
	c[j] = r2*0.5*j*(r+sigma_sqr*j);
    };
    vector<double> f_next(M+1);
    for (int m=0;m<=M;++m) { f_next[m]=max(0.0,K-S_values[m]); };
    vector<double> f(M+1);
    for (int t=N-1;t>=0;--t) {
	f[0]=K;
	for (int m=1;m<M;++m) {
	    f[m]=a[m]*f_next[m-1]+b[m]*f_next[m]+c[m]*f_next[m+1];
	    f[m] = max(f[m],K-S_values[m]);  // check for exercise
	};
	f[M] = 0;
	for (int m=0;m<=M;++m) { f_next[m] = f[m]; };
    };
    return f[M/2];
};

 
void Progonka_in_place(const  std::vector<double>& low, 
					   const  std::vector<double>& diag, const  std::vector<double>& up,  std::vector<double>& rightside_result)
{
      int N = diag.size();
      if((low.size()!=N-1)||(low.size()!=N-1)||(rightside_result.size()!=N)) {
		  std::cout << "les tailles des vecteurs sont incorrectes!" << std::endl;
		  std::exit(1);
      }

      std::vector<double> A(N-1), B(N-1);

      // descente
      const std::vector<double> & rightside = rightside_result;
      A[0] = -up[0]/diag[0];
      B[0] = rightside[0]/diag[0];
      double denom;
      for(int i=1;i<N-1;i++)
      {
        denom = diag[i] + low[i-1]*A[i-1];
        A[i] = -up[i]/denom;
        B[i] = (rightside[i]-low[i-1]*B[i-1])/denom;
      };

      // remontee
      std::vector<double> & result = rightside_result;
      result[N-1] = (rightside[N-1] - low[N-2]*B[N-2])/(diag[N-1] + low[N-2]*A[N-2]);
      for(int i=N-2;i>=0;i--)
      {
        result[i] = A[i]*result[i+1] + B[i];
      };
}



void Progonka( std::vector<double>::const_iterator low_begin,  
			  std::vector<double>::const_iterator diag_begin,  std::vector<double>::const_iterator up_begin, 
			std::vector<double>::const_iterator rs_begin,  std::vector<double>::iterator result_begin,  std::vector<double>::iterator result_end)
{
      int N = result_end - result_begin;

      std::vector<double> A(N-1), B(N-1);

      // descente
       std::vector<double>::const_iterator low_iter = low_begin;
       std::vector<double>::const_iterator diag_iter = diag_begin;
       std::vector<double>::const_iterator up_iter = up_begin;
       std::vector<double>::const_iterator rs_iter = rs_begin;


      A[0] = -(*up_iter)/(*diag_iter);
      B[0] = (*rs_iter)/(*diag_iter);
      double denom;
      for(int i=1;i<N-1;i++)
      {
        diag_iter++;
        up_iter++;
        rs_iter++;

        denom = (*diag_iter) + (*low_iter)*A[i-1];
        A[i] = -(*up_iter)/denom;
        B[i] = ((*rs_iter) - (*low_iter)*B[i-1])/denom;

        low_iter++;
      };


      // remontee
       std::vector<double>::iterator result_iter = --result_end;
      diag_iter++;
      rs_iter++;
      *result_iter = ((*rs_iter) - (*low_iter)*B[N-2])/((*diag_iter) + (*low_iter)*A[N-2]);
      result_iter--;

      for(int i = N - 2; i >= 0; i--)
      {
         *(result_iter) = A[i]*(*(result_iter+1)) + B[i];
         result_iter--;
      };
  }



 double option_price_put_american_finite_diff_implicit(const double& S,
							   const double& K, 
							   const double& r,
							   const double& sigma,
							   const double& time, 
							   const int& no_S_steps,
							   const int& no_t_steps) {
    double sigma_sqr = sigma*sigma;
    const int M=no_S_steps + (no_S_steps%2); // need no_S_steps to be even:

    double delta_S = 2.0*S/double(M);
     int N=no_t_steps;
    double delta_t = time/N;
    
	std::vector<double> low(M);      // diagonale inferieure
	std::vector<double> diag(M + 1); // diagonale principale
	std::vector<double> up(M);       // diagonale superieure
	std::vector<double> Price(M + 1);
	std::vector<double> S_values(M + 1);

    for (int m=0;m<=M;m++) { S_values[m] = m*delta_S; };

	// conditions au bord: propagation des valeurs initiales:
	up[0] = 0;
	diag[0] = 1;
	low[M - 1] = 0;
	diag[M] = 1;
 
    for (int j=1;j<M;++j) {
		low[j-1] = 0.5*j*delta_t*(r-sigma_sqr*j);    // a[j]
		diag[j]   = 1.0 + delta_t*(r+sigma_sqr*j*j);  // b[j];
		up[j] = 0.5*j*delta_t*(-r-sigma_sqr*j);   // c[j];
    };
 
 
    for (int m=0;m<=M;++m){  Price[m] = max(0.0,K-S_values[m]); };

    for(int t=N;t>0;--t) {
		Progonka_in_place(low, diag, up, Price);
	for (int m=1;m<M;++m) { 	// now check for exercise
	    Price[m] = max( Price[m], K-S_values[m]);
	};
    };
	return Price[M/2];
};


//double option_price_put_american_finite_diff_implicit_itpp(const double& S,
//							   const double& K, 
//							   const double& r,
//							   const double& sigma,
//							   const double& time, 
//							   const int& no_S_steps,
//							   const int& no_t_steps) {
//    double sigma_sqr = sigma*sigma;
//    int M=no_S_steps + (no_S_steps%2); // need no_S_steps to be even:
//
//    double delta_S = 2.0*S/double(M);
//    double S_values[M+1];
//    for (int m=0;m<=M;m++) { S_values[m] = m*delta_S; };
//    int N=no_t_steps;
//    double delta_t = time/N;
//    
//    mat A = zeros(M+1,M+1);
//    A(0,0) = 1.0;
//    for (int j=1;j<M;++j) {
//	A(j,j-1) = 0.5*j*delta_t*(r-sigma_sqr*j);    // a[j]
//	A(j,j)   = 1.0 + delta_t*(r+sigma_sqr*j*j);  // b[j];
//	A(j,j+1) = 0.5*j*delta_t*(-r-sigma_sqr*j);   // c[j];
//    };
//    A(M,M)=1.0;
//    vec B(M+1);
//    for (int m=0;m<=M;++m){ B(m) = max(0.0,K-S_values[m]); };
//    mat InvA = inv(A);
//    vec F=InvA*B;
//    for(int t=N-1;t>0;--t) {
//	B = F;
//	F = InvA*B;
//	for (int m=1;m<M;++m) { 	// now check for exercise
//	    F(m) = max(F(m), K-S_values[m]);
//	};
//    };
//    return F(M/2);
//};

double option_price_call_european_binomial( const double& S,     // spot price
					    const double& K,     // exercice price
					    const double& r,     // interest rate
					    const double& sigma, // volatility
					    const double& t,     // time to maturity
					    const int& steps){  // no steps in binomial tree
   double R = exp(r*(t/steps));            // interest rate for each step
   double Rinv = 1.0/R;                    // inverse of interest rate
   double u = exp(sigma*sqrt(t/steps));    // up movement
   double uu = u*u;
   double d = 1.0/u;
   double p_up = (R-d)/(u-d);
   double p_down = 1.0-p_up;
   vector<double> prices(steps+1);       // price of underlying
   prices[0] = S*pow(d, steps);  // fill in the endnodes.
   for (int i=1; i<=steps; ++i) prices[i] = uu*prices[i-1];
   vector<double> call_values(steps+1);       // value of corresponding call 
   for (int i=0; i<=steps; ++i) call_values[i] = max(0.0, (prices[i]-K)); // call payoffs at maturity
   for (int step=steps-1; step>=0; --step) {
      for (int i=0; i<=step; ++i) {
	 call_values[i] = (p_up*call_values[i+1]+p_down*call_values[i])*Rinv;
      };
   };
   return call_values[0];
};

double option_price_put_american_trinomial( const double& S,  
					    const double& K,  
					    const double& r,  
					    const double& q,
					    const double& sigma,
					    const double& t,    
					    const int& steps) { 
    double delta_t = t/steps;
    double Rinv = exp(-r*(delta_t));
    double sigma_sqr=pow(sigma,2);

    double u = exp(sigma*sqrt(3.0*delta_t)); 
    double d = 1.0/u;
    double p_u = 1.0/6.0 + sqrt(delta_t/(12.0*sigma_sqr)) * (r-q-0.5*sigma_sqr);
    double p_m = 2.0/3.0;
    double p_d = 1.0/6.0 - sqrt(delta_t/(12.0*sigma_sqr)) * (r-q-0.5*sigma_sqr);

    vector< vector<double> > Stree;       // price of underlying in a tree
    vector<double> Svec;
    Svec.push_back(S);
    for (int step=1;step<=steps;++step){ 
	Stree.push_back(Svec);
	Svec.insert(Svec.begin(),Svec[0]*d); // use the fact that only the extreme values change. 
	Svec.push_back(Svec[Svec.size()-1]*u);
    };
    int m = Svec.size();
    vector<double> values_next = vector<double>(m);       // value of option next step
    for (int i=0; i<m; ++i) values_next[i] = max(0.0, K-Svec[i]); // call payoffs at maturity
    vector<double> values;
    for (int step=steps-1; step>=0; --step) {
	m = Stree[step].size();	
	values = vector<double> (m);       // value of option
	for (int i=0; i<m; ++i) {
	    values[i] = (p_u*values_next[i+2]+p_m*values_next[i+1] + p_d*values_next[i])*Rinv;
	    values[i] = max(values[i],K-Stree[step][i]);       // check for exercise
	};
	values_next=values;
    };
    return values[0];
};

